/*
* This code isdistrubuted under GPLv3
* 
* This code solves the 2D Poasson problem u_{xx} + u_{yy} = f, over [0,1]^2 with Dirichlet boundary conditions,
* 
* This code implements Finite Element method using trinagular elements with uniform mesh,
* piece wise linear basis functions and GMRES iterative solver for the linear system.
* 
* GMRES is accelerated with MILU preconditioner:
* 
* 
* Code Author: Miroslav Stoyanov, Oct 2011
* 
* Copyright (C) 2011  Miroslav Stoyanov
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* Since the GNU General Public License is longer than this entire code, 
* a copy of it can be obtained separately at <http://www.gnu.org/licenses/>
* 
*/

#include "mesh_quadrature.h"

double boundary( double x, double y ){
	if ( x < 1.E-8 ){ return 0.0; };
	if ( x > 1.0-1.E-8 ){ return 0.0; };
	if ( y < 1.E-8 ){ return sin( M_PI * x ); };
	if ( y > 1.0-1.E-8 ){ return -sin( M_PI *x ); };
};

double u_exact( double x, double y ){
	return sin( M_PI * x ) * cos( M_PI * y );
};

double f( double x, double y ){
	return -2.0 * M_PI * M_PI * sin( M_PI * x ) * cos( M_PI * y );
};

// GMRES Routines
struct GMRESData{
	struct Pmat *A;
	struct FactorizedMILU *iLU;
};

void funcA( double *x, double *r, void *ud ){
	struct GMRESData *gd = (struct GMRESData*) ud;
	pmat_act( *(gd->A), x, r, 1.0 );
	MILU_solve( gd->iLU, r );
};


int main( int argc, char **argv ){
        int i;
        int iDirectSolver = 1;
        if ( argc == 1 ){
                printf(" Please specify the number of intervals to use\n");
                printf(" 	i.e. assignment4 100\n");
                return 1;
        };
	if ( argc == 3 ){
		iDirectSolver = 0;
	};

        
        int N = atoi( argv[1] );
        double dx = 1.0 / ( (double) N );
        
        struct GeometryMesh mesh;
        
        GeometryMeshFormUnit( &mesh, dx, dx );
        printf(" Mesh formed with %d elements and %d degrees of freedom.\n",mesh.num_elem,mesh.num_u);
        
        struct Pmat L;
        double *R = (double*) malloc( mesh.num_u * sizeof( double ) );
        double *u = (double*) malloc( mesh.num_u * sizeof( double ) );
        
        GeometryFormLHS( mesh, &L ); // form the left hand side matrix
        
        GeometryFromRHS( mesh, R, &f, &boundary ); // form the right-hand side vector
        
        if ( iDirectSolver == 1 ){
		// solve the system usnig SuiteSparse
		printf(" Calling UMFPACK Solver\n");
		pmat_UMFPACK( L, R, u );
	}else{
	        // solve the system usnig GMRES
	        struct FactorizedMILU iLU;
	        
	        // form the preconditioner
	        MILU_analyze( L, &iLU );        
		MILU_factorize( L, &iLU );
	        
		// set the preconditioned RHS
		MILU_solve( &iLU, R );
		
		// start with an initial guess of all zeros
		for( i=0; i<mesh.num_u; i++ ){
			u[i] = 0;
		};
	
	
		// set the data for the GMRES problem
		struct GMRESData ud;
		ud.A = &L;
		ud.iLU = &iLU;
		
	        // call GMRES
		if ( GMRES( &funcA, u, R, &ud, mesh.num_u, 30, 50, 1.E-9 ) == 1 ){
			printf(" Warning: the GMRES method failed to converge.\n");
		};
		MILU_free( &iLU );
	};
	
        
	// compute the error
	GeometryLoadSolution( &mesh, u, &boundary );
	
	double err = GeometryComputerErrorL2( mesh, &u_exact );
	
	printf(" The L2 error for the problem is: %e\n",err );
	
	//GeometrySaveSolution( mesh, "finalsolution" );
        
        // clean up after ourselves
        free( R );
        free( u );
        pmat_free( &L );
        GeometryMeshFree( &mesh );
        
        return 0;
};
