#ifndef __SPARSE_MATRIX_C
#define __SPARSE_MATRIX_C

/*
* This code isdistrubuted under GPLv3
* 
* Code Author: Miroslav Stoyanov, Oct 2011
* 
* Copyright (C) 2011  Miroslav Stoyanov
* 
* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* Since the GNU General Public License is longer than this entire code, 
* a copy of it can be obtained separately at <http://www.gnu.org/licenses/>
* 
*/

#include "sparse_matrix.h"

void pattern_init( struct Pattern *p, int n ){
	int i;
	p->sp = (struct Pnode**) malloc( n * sizeof( struct Pnode* ) );
	p->count = (int*) malloc( n * sizeof(int) );
	p->num_rows = n;
	for( i=0; i<n; i++ ){ p->count[i] = 0; p->sp[i] = NULL; };
};

void pattern_add( struct Pattern *p, int r, int indx ){
	if ( p->count[r] == 0 ){ // add as the first entry
		p->sp[r] = (struct Pnode*) malloc( sizeof(struct Pnode) );
		p->sp[r] -> next = NULL;
		p->sp[r] -> indx = indx;
		p->count[r]++;
	}else if ( p->sp[r] -> indx > indx ){ // add before the first entry
		struct Pnode *t = (struct Pnode*) malloc( sizeof(struct Pnode) );
		t -> indx = indx;
		t -> next = p->sp[r];
		p->sp[r] = t;
		p->count[r]++;
	}else if ( p->sp[r] -> indx == indx ){ // do nothing (entrie is repeating)
	}else if ( p->count[r] == 1 ){ // only one node, add after the node
		p->sp[r] -> next = (struct Pnode*) malloc( sizeof(struct Pnode) );
		p->sp[r] -> next -> indx = indx;
		p->sp[r] -> next -> next = NULL;
		p->count[r]++;
	}else{ // we have at least two nodes, then we can loop looking one node ahead
		struct Pnode *t = p->sp[r];
		int loop = 1;
		while ( loop == 1 ){
			if ( t -> next -> indx == indx ){ // entrie already here
				loop = 0;
			}else if ( t -> next -> indx > indx ){ // add here
				struct Pnode *d = (struct Pnode*) malloc( sizeof(struct Pnode) );
				d -> next = t -> next;
				t -> next = d;
				d -> indx = indx;
				p->count[r]++;
				loop = 0;
			}else if ( t -> next -> next == NULL ){ // if next entrie is still smaller than the index
				t = t -> next;
				t -> next = (struct Pnode*) malloc( sizeof(struct Pnode) );
				t = t -> next;
				t -> next = NULL;
				t -> indx = indx;
				p->count[r]++;
				loop = 0;
			}else{ // go on
				t = t -> next;
			};
		};
	};
};

void pattern_del( struct Pattern *p ){
	int i;
	struct Pnode *t;
	for ( i=0; i<p->num_rows; i++ ){
		t = p->sp[i];
		while ( p->sp[i] != NULL ){
			p->sp[i] = p->sp[i] -> next;
			free( t );
			t = p->sp[i];
		};
	};
	free( p->sp );
	free( p->count );
	p->num_rows = 0;
};



void pmat_create( struct Pmat *m, struct Pattern p ){
	int i, j, count = 0;

	// set the constants
	m->num_rows = p.num_rows;

	// count number of entries
	for( i=0; i<p.num_rows; i++ ){
		count += p.count[i];
	};

	// declare the memory needed
	m -> pntr = (int*) malloc( (m->num_rows + 1) * sizeof(int) );
	m -> indx = (int*) malloc( count * sizeof(int) );
	m -> vals = (double*) malloc( count * sizeof(double) );

	// set the pointers
	m->pntr[0] = 0;
	for( i=1; i<m->num_rows; i++ ){
		m->pntr[i] = m->pntr[i-1] + p.count[i-1];
	};
	m->pntr[m->num_rows] = count;

	// copy the indexes
	struct Pnode * t;
	for( i=0; i<m->num_rows; i++ ){
		t = p.sp[i];
		for( j=m->pntr[i]; j<m->pntr[i+1]; j++ ){
			m->indx[j] = t -> indx;
			t = t -> next;
		};
	};

	//dscal( count, 0.0, m->vals, 1 ); // this is buggy sometimes
        for( i=0; i<count; i++ ){ m->vals[i] = 0.0; };
};

void pmat_delete( struct Pmat *m ){
	free( m->pntr ); m->pntr = NULL;
	free( m->indx ); m->indx = NULL;
	free( m->vals ); m->vals = NULL;
	m->num_rows = 0.0;
};

void pmat_addval( struct Pmat *m, int r, int c, double val ){
	int start = m->pntr[r];
	int end = m->pntr[r+1];
	int i = (start+end)/2;
	while ( m->indx[i] != c ){
		if ( m->indx[i] < c ){
			start = i + 1;
		}else{
			end = i - 1;
		};
		i = (start+end)/2;
	};
	m->vals[i] += val;
};

void pmat_repval( struct Pmat *m, int r, int c, double val ){
	int start = m->pntr[r];
	int end = m->pntr[r+1];
	int i = (start+end)/2;
	while ( m->indx[i] != c ){
		if ( m->indx[i] < c ){
			start = i + 1;
		}else{
			end = i - 1;
		};
		i = (start+end)/2;
	};
	m->vals[i] += val;
};

int pmat_test_index( struct Pmat *m, int r, int c ){
        if ( (r<0) || (r>m->num_rows) ){ return 1; }; // the row is out of range
        int i;
        for( i=m->pntr[r]; i<m->pntr[r+1]; i++ ){
                if ( m->indx[i] == c ){ return 0; }; // found the entry
        };
        return 1; // looped over the row and the entry is missing
};

void pmat_fread( FILE *ifs, struct Pmat *m ){
        int read;
	read = fread( &(m->num_rows), 1, sizeof(int), ifs );
	
	m->pntr = (int*) malloc( (m->num_rows)*sizeof(int) );
	read = fread( m->pntr, m->num_rows+1, sizeof(int), ifs );

	m->indx = (int*) malloc( m->pntr[m->num_rows] * sizeof(int) );
	m->vals = (double*) malloc( m->pntr[m->num_rows] * sizeof(double) );
	read = fread( m->indx, m->pntr[m->num_rows], sizeof(int), ifs );
	read = fread( m->vals, m->pntr[m->num_rows], sizeof(double), ifs );
};

void pmat_fwrite_ascii( struct Pmat m, const char *filename ){ // writes m into an ascii file
        int i;
        FILE * fout = fopen( filename, "w");
        fprintf(fout,"%d \n",m.num_rows);
        for( i=0; i<m.num_rows+1; i++ ){
                fprintf(fout,"%d\n",m.pntr[i]);
        };
        for( i=0; i<m.pntr[m.num_rows]; i++ ){
                fprintf(fout,"%d\n",m.indx[i]);
        };
        for( i=0; i<m.pntr[m.num_rows]; i++ ){
                fprintf(fout,"%2.16lf\n",m.vals[i]);
        };
        fclose( fout );
};

void pmat_act( struct Pmat m, double *x, double *r, double alpha ){
	int i;
        for( i=0; i < m.num_rows; i++ ){
                r[i] = alpha * ddoti( m.pntr[i+1] - m.pntr[i], &(m.vals[m.pntr[i]]), &(m.indx[m.pntr[i]]), &x[1] );
        };
};

void pmat_free( struct Pmat *m ){
	free( m->pntr );
	free( m->indx );
	free( m->vals );
	m->num_rows = 0;
};

void MILU_analyze( struct Pmat m, struct FactorizedMILU *ilu ){
	int i, j, k;
	ilu->num_rows = m.num_rows;
	ilu->pntr = (int*) malloc( (ilu->num_rows+1) * sizeof(int) );
	ilu->pntrC = (int*) malloc( (ilu->num_rows+1) * sizeof(int) );
	ilu->indxD = (int*) malloc( ilu->num_rows * sizeof(int) );
	for( i=0; i <= ilu->num_rows; i++ ){
		ilu->pntr[i] = m.pntr[i];
	};
	ilu->pntrC[ilu->num_rows] = ilu->pntr[ilu->num_rows];
	ilu->indxT = (int*) malloc( ilu->pntr[ilu->num_rows] * sizeof(int) );
	ilu->indx = (int*) malloc( ilu->pntr[ilu->num_rows] * sizeof(int) );
	ilu->indxC = (int*) malloc( ilu->pntr[ilu->num_rows] * sizeof(int) );
	ilu->vals = (double*) malloc( ilu->pntr[ilu->num_rows] * sizeof(double) );
	for( i=0; i < ilu->pntr[ilu->num_rows]; i++ ){
		ilu->indx[i] = m.indx[i];
	};

	// get the diagonal entrie for each row
	for( i=0; i<ilu->num_rows; i++ ){
		j = ilu->pntr[i];
		while( (j<ilu->pntr[i+1])&&(ilu->indx[j]<i) ){ j++; }
		if ( ilu->indx[j] == i ){ ilu->indxD[i] = j; };
	};

	for( i=0; i<ilu->num_rows; i++ ){
		ilu->pntrC[i] = 0;
	};
	for( j=0; j<ilu->pntr[ilu->num_rows]; j++ ){
		ilu->pntrC[ ilu->indx[j] + 1 ]++;
	};
	for( i=1; i<=ilu->num_rows; i++ ){
		ilu->pntrC[i] += ilu->pntrC[i-1];
	};

	int *cnt = (int*) malloc( ilu->num_rows * sizeof(int) );
	for( i=0; i<ilu->num_rows; i++ ){
		cnt[i] = 0;
	};

	for( i=0; i<ilu->num_rows; i++ ){
		for( j=ilu->pntr[i]; j<ilu->pntr[i+1]; j++ ){
			ilu->indxC[ ilu->pntrC[ ilu->indx[j] ] + cnt[ ilu->indx[j] ] ] = i;
			ilu->indxT[ ilu->pntrC[ ilu->indx[j] ] + cnt[ ilu->indx[j] ] ] = j;
			cnt[ ilu->indx[j] ] ++;
		};
	};

	free( cnt );
};

void MILU_factorize( struct Pmat m, struct FactorizedMILU *ilu ){
	int N, i, j, jr, jt, k, ii, jj;
	int start, end, current;
	int found;
	double d, e;
	
	N = m.num_rows;

	dcopy( m.pntr[N], m.vals, 1, ilu->vals, 1 );

	for( i=0; i<N-1; i++ ){
		d = 1.0 / ilu->vals[ ilu->indxD[i] ];
		for( j=ilu->indxD[i]+1; j<ilu->pntrC[i+1]; j++ ){
			jr = ilu->indxC[j];
			jt = ilu->indxT[j];
			e = d* ilu->vals[jt];
			ilu->vals[jt] = e;
			for( k=ilu->indxD[i]+1; k < ilu->pntr[i+1]; k++ ){
				start = ilu->pntr[jr];
				end = ilu->pntr[jr+1] - 1;
				current = (start+end) / 2;
				while( (start<end) && (ilu->indx[current] != ilu->indx[k]) ){
					if ( ilu->indx[current] < ilu->indx[k] ){
						start = current + 1;
					}else{
						end = current - 1;
					};
					current = (start+end)/2;
				};
				if ( ilu->indx[current] == ilu->indx[k] ){
					ilu->vals[current] -= e * ilu->vals[k];
				}else{
					ilu->vals[ilu->indxD[jr]] -= e * ilu->vals[k];
				};
			};
		};
	};
};

void MILU_solve( struct FactorizedMILU *ilu, double *x ){
	int i;
	for( i=1; i < ilu->num_rows; i++ ){
		if ( (ilu->indxD[i] - ilu->pntr[i]) > 0 )
			x[i] -= ddoti( (ilu->indxD[i] - ilu->pntr[i]), &(ilu->vals[ilu->pntr[i]]), &(ilu->indx[ilu->pntr[i]]), &x[1] );
	};
	for( i=ilu->num_rows-1; i>=0; i-- ){
		if ( (ilu->pntr[i+1] - ilu->indxD[i] - 1) > 0 ){
			x[i] -= ddoti( (ilu->pntr[i+1] - ilu->indxD[i] - 1), &(ilu->vals[ilu->indxD[i]+1]), &(ilu->indx[ilu->indxD[i]+1]), &x[1] );
		};
		x[i] /= ilu->vals[ilu->indxD[i]];
	};
};

void MILU_free( struct FactorizedMILU *ilu ){
	free( ilu->pntr );
	free( ilu->indx );
	free( ilu->vals );
	free( ilu->indxC );
	free( ilu->indxD );
	free( ilu->indxT );
};

int GMRES( void (*funcA)(double*,double*,void*), double *x, double *b, void *ud, int size, int max_inner, int max_outer, double tol ){
	int i, j; // loop variables i, j and k is the size of the computed basis
        if ( max_inner < 2 ){ max_inner = 2; }; // it makes no sense to stop before 2 iterations (unless we have converged)
        double H[max_inner][max_inner+1]; // holds the transformation for the normalized basis
        
        // the total size of the W array can be too big, so we need to delcare it with malloc as opposed to []
        double *W[max_inner+1]; // work space for the vector basis
        for( i=0; i<max_inner+1; i++ ){ W[i] = (double*) malloc( size*sizeof( double ) ); };
        
        double S[max_inner], C[max_inner+1]; // sin and cos of the Givens rotations
        double Z[max_inner]; // holds the coefficients of the solution
        
        
        double alpha, h_k; // temp variables
        
        double outer_res = tol + 1.0, inner_res; // outer and inner residual
        int outer_itr = 0, inner_itr; // counts the inner and outer iterations
        
        while ( (outer_res > tol ) && ( outer_itr < max_outer ) ){
                
                funcA( x, W[1], ud ); // call the funtion to compute W_1 = A x
                
                dcopy( size, b, 1, W[0], 1 );
                daxpy( size, -1.0, W[1], 1, W[0], 1 ); // W_0 = b - Ax
                
                Z[0] = dnrm2( size, W[0], 1 );
                dscal( size, 1.0/Z[0], W[0], 1 ); // save the first normalized vector
                
                inner_res = Z[0]; // first residual
                inner_itr = 0; // counts the size of the basis
                
                while ( (inner_res > tol) && (inner_itr < max_inner) ){
                        inner_itr++;
                        
                        funcA( W[inner_itr-1], W[inner_itr], ud );
                        
                        for( i=0; i<inner_itr; i++ ){ // normalize the vector with respect to the previous vectors
                                H[inner_itr-1][i] = ddot( size, W[inner_itr], 1, W[i], 1 );
                                daxpy( size, -H[inner_itr-1][i], W[i], 1, W[inner_itr], 1 );
                        };
                        
                        h_k = dnrm2( size, W[inner_itr], 1 ); // set the next basis vector
                        dscal( size, 1.0/h_k, W[inner_itr], 1 );
                        
                        for ( i=0; i<inner_itr-1; i++ ){ // form the next row of the transformation
                                alpha = H[inner_itr-1][i];
                                H[inner_itr-1][i]   = C[i] * alpha + S[i] * H[inner_itr-1][i+1];
                                H[inner_itr-1][i+1] = S[i] * alpha - C[i] * H[inner_itr-1][i+1];
                        };
                        
                        alpha = sqrt( h_k * h_k  +  H[inner_itr-1][inner_itr-1] * H[inner_itr-1][inner_itr-1] );
                        
                        // set the next set of Givens rotations
                        S[inner_itr-1] = h_k / alpha;	        
                        C[inner_itr-1] = H[inner_itr-1][inner_itr-1] / alpha;
			
                        H[inner_itr-1][inner_itr-1] = alpha;
                        
                        // Z is used to reconstruct the solution in the end
                        Z[inner_itr] = S[inner_itr-1]*Z[inner_itr-1];
			Z[inner_itr-1] = C[inner_itr-1]*Z[inner_itr-1]; // apply it on z
                        
                        inner_res = fabs(Z[inner_itr]);
                };
                
                inner_itr--;
                
                if ( inner_itr > -1 ){ // if the first guess was not within TOL of the true solution
			Z[inner_itr] /= H[inner_itr][inner_itr];
			for( i=inner_itr-1; i>-1; i-- ){
				h_k = 0.0;
				for( j=i+1; j<=inner_itr; j++ ){
					h_k += H[j][i] * Z[j];
				};
				Z[i] = ( Z[i] - h_k ) / H[i][i];
			};

			for( i=0; i<=inner_itr; i++ ){
				daxpy( size, Z[i], W[i], 1, x, 1 );
			};
		};
                
                outer_res = inner_res;
                outer_itr++;
        };
        
        for( i=0; i<max_inner+1; i++ ){ free( W[i] ); };
        printf(" GMRES Finished in %d outer plus %d inner iterations with final residual %e\n",outer_itr,inner_itr+1,outer_res);
        return (outer_res > tol) ? 1 : 0;
};

// SuiteSparse
void pmat_UMFPACK_factorize( struct Pmat m, struct FactorizedUMFPACK *umf ){
	int i;
	umf->num_rows = m.num_rows;
	umf->pntr = (int*) malloc( ((umf->num_rows+1) + (umf->num_rows+1) % 2) * sizeof(int) );
	for( i=0; i<=umf->num_rows; i++ ){
		umf->pntr[i] = m.pntr[i];
	};
	umf->indx = (int*) malloc( (umf->pntr[umf->num_rows] + umf->pntr[umf->num_rows]%2) * sizeof(int) );
	umf->vals = (double*) malloc( umf->pntr[umf->num_rows] * sizeof(double) );
	for( i=0; i<umf->pntr[umf->num_rows]; i++ ){
		umf->indx[i] = m.indx[i];
	};
	dcopy( umf->pntr[umf->num_rows], m.vals, 1, umf->vals, 1 );

	void *Symbolic;
	umfpack_di_symbolic( umf->num_rows, umf->num_rows, umf->pntr, umf->indx, umf->vals, &Symbolic, NULL, NULL);
	umfpack_di_numeric( umf->pntr, umf->indx, umf->vals, Symbolic, &(umf->Numeric), NULL, NULL);
	umfpack_di_free_symbolic(&Symbolic);
};

void pmat_UMFPACK_solve( struct FactorizedUMFPACK umf, double *b, double *x ){
	umfpack_di_solve(UMFPACK_A, umf.pntr, umf.indx, umf.vals, x, b, umf.Numeric, NULL, NULL);
};

void pmat_UMFPACK_free( struct FactorizedUMFPACK *umf ){
	free( umf->pntr ); umf->pntr = NULL;
	free( umf->indx ); umf->indx = NULL;
	free( umf->vals ); umf->vals = NULL;
	umfpack_di_free_numeric( &(umf->Numeric) );
	umf->num_rows = 0;
};

void pmat_UMFPACK( struct Pmat m, double *b, double *x ){
	struct FactorizedUMFPACK umf;
	pmat_UMFPACK_factorize( m, &umf );
	pmat_UMFPACK_solve( umf, b, x );
	pmat_UMFPACK_free( &umf );
};





///////////////////////////
// --- BLAS ROUTINES --- //
///////////////////////////
// BLAS 1
double ddot( int N, double *a, int inca, double *b, int incb ){
	return ddot_( &N, a, &inca, b, &incb );
};
double dnrm2( int N, double *a, int inca ){
	return dnrm2_( &N, a, &inca );
};
void dcopy(  int N, double *a, int inca, double *b, int incb ){
	dcopy_( &N, a, &inca, b, &incb );
};
void daxpy( int N, double alpha, double *a, int inca, double *b, int incb ){
	daxpy_( &N, &alpha, a, &inca, b, &incb );
};
void dscal( int N, double alpha, double *a, int inca ){
	dscal_( &N, &alpha, a, &inca );
};


double ddoti( int N, double *a, int *indx, double *x ){
        int i;
        double sum = 0.0;
        for( i=0; i<N; i++ ){
                sum += a[i] * x[ indx[i] - 1];
        };
        return sum;
};

#endif
